def create_minor_y_ticks(ylim):
    import numpy as np

    diff = abs(ylim[1] - ylim[0])
    if diff > 100:
        inc = 10
    if diff > 20:
        inc = 5
    elif diff > 10:
        inc = 2.5
    elif diff > 2:
        inc = 1
    else:
        inc = 0.25
    print(ylim)
    if ylim[0] > 2:
        lower_bound = int(ylim[0])
    else:
        lower_bound = ylim[0]
    while lower_bound % inc != 0:
        lower_bound -= 1
    if ylim[1] > 2:
        upper_bound = int(ylim[1])
    else:
        upper_bound = ylim[1]
    while upper_bound % inc != 0:
        upper_bound += 1
    upper_bound += inc
    print(f"{lower_bound = }, {upper_bound = }, {inc = }")
    major_yticks = np.arange(lower_bound, upper_bound, inc)
    minor_yticks = np.arange(lower_bound, upper_bound, inc/2)[:-1]
    return major_yticks, minor_yticks


def violin_plot(
    df,
    df_labels_and_columns: {},
    output_filename: str,
    plt_title: str = None,
    bottom: float = 0.4,
    ylim: list = None,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlable=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    ylabel=r"Error ($\mathrm{kcal\cdot mol^{-1}}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
) -> None:
    """
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd

    print(f"Plotting {output_filename}")
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    vLabels, vData = [], []
    annotations = []  # [(x, y, text), ...]
    cnt = 1
    plt.rcParams["text.usetex"] = usetex
    for k, v in df_labels_and_columns.items():
        df[v] = pd.to_numeric(df[v])
        df_sub = df[df[v].notna()].copy()
        vData.append(df_sub[v].to_list())
        if usetex:
            k_label = "\\textbf{" + k + "}"
        else:
            k_label = k
        k_label = convert_deltas_ssapt0(k_label)
        vLabels.append(k_label)
        m = df_sub[v].max()
        rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
        mae = df_sub[v].apply(lambda x: abs(x)).mean()
        max_error = df_sub[v].apply(lambda x: abs(x)).max()
        if usetex:
            text = r"\textit{%.2f}" % mae
            text += "\n"
            text += r"\textbf{%.2f}" % rmse
            text += "\n"
            text += r"\textrm{%.2f}" % max_error
        else:
            text = r"%.2f" % mae
            text += "\n"
            text += r"%.2f" % rmse
            text += "\n"
            text += r"%.2f" % max_error
        annotations.append((cnt, m, text))
        cnt += 1

    pd.set_option("display.max_columns", None)
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    ax = plt.subplot(111)
    vplot = ax.violinplot(
        vData,
        showmeans=True,
        showmedians=False,
        quantiles=[[0.05, 0.95] for i in range(len(vData))],
        widths=widths,
    )
    for n, partname in enumerate(["cbars", "cmins", "cmaxes", "cmeans"]):
        vp = vplot[partname]
        vp.set_edgecolor("black")
        vp.set_linewidth(1)
        vp.set_alpha(1)
    quantile_color = "red"
    quantile_style = "-"
    quantile_linewidth = 0.8
    for n, partname in enumerate(["cquantiles"]):
        vp = vplot[partname]
        vp.set_edgecolor(quantile_color)
        vp.set_linewidth(quantile_linewidth)
        vp.set_linestyle(quantile_style)
        vp.set_alpha(1)

    colors = ["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))]
    for n, pc in enumerate(vplot["bodies"], 1):
        pc.set_facecolor(colors[n - 1])
        pc.set_alpha(0.6)

    vLabels.insert(0, "")
    xs = [i for i in range(len(vLabels))]
    xs_error = [i for i in range(-1, len(vLabels) + 1)]
    x_error_label = r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$" if usetex else r"within 1 kcal/mol"
    ax.plot(
        xs_error,
        [1 for i in range(len(xs_error))],
        "k--",
        label=x_error_label,
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        xs_error,
        [0 for i in range(len(xs_error))],
        "k--",
        linewidth=0.5,
        alpha=0.5,
        # label=r"Reference Energy",
        zorder=0,
    )
    ax.plot(
        xs_error,
        [-1 for i in range(len(xs_error))],
        "k--",
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        [],
        [],
        linestyle=quantile_style,
        color=quantile_color,
        linewidth=quantile_linewidth,
        label=r"5-95th Percentile",
    )
    navy_blue = (0.0, 0.32, 0.96)
    ax.set_xticks(xs)
    plt.setp(
        ax.set_xticklabels(vLabels),
        rotation=x_label_rotation,
        fontsize=x_label_fontsize,
    )
    ax.set_xlim((0, len(vLabels)))
    if ylim is not None:
        ax.set_ylim(ylim)
        major_yticks, minor_yticks = create_minor_y_ticks(ylim)
        ax.set_yticks(major_yticks)
        ax.set_yticks(minor_yticks, minor=True)

    lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

    if set_xlable:
        ax.set_xlabel("Level of Theory", color="k")
    ax.set_ylabel(ylabel, color="k")

    ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
    # Annotations of RMSE
    for x, y, text in annotations:
        ax.annotate(
            text,
            xy=(x, y),
            xytext=(x, y + 0.1),
            color="black",
            fontsize="8",
            horizontalalignment="center",
            verticalalignment="bottom",
        )

    for n, xtick in enumerate(ax.get_xticklabels()):
        xtick.set_color(colors[n - 1])
        xtick.set_alpha(0.8)

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
        path = f"{output_basename}_violin.{ext}"
    else:
        path = output_filename
    print(f"{path}")
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    return


def violin_plot_table(
    df,
    df_labels_and_columns: {},
    output_filename: str,
    plt_title: str = None,
    bottom: float = 0.4,
    ylim: list = None,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlabel=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    table_fontsize=8,
    ylabel=r"Error (kcal$\cdot$mol$^{-1}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
) -> None:
    """
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import gridspec

    print(f"Plotting {output_filename}")
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    vLabels, vData = [], []
    annotations = []  # [(x, y, text), ...]
    cnt = 1
    plt.rcParams["text.usetex"] = usetex
    for k, v in df_labels_and_columns.items():
        df[v] = pd.to_numeric(df[v])
        df_sub = df[df[v].notna()].copy()
        vData.append(df_sub[v].to_list())
        k_label = "\\textbf{" + k + "}"
        k_label = convert_deltas_ssapt0(k_label)
        vLabels.append(k_label)
        m = df_sub[v].max()
        rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
        mae = df_sub[v].apply(lambda x: abs(x)).mean()
        max_pos_error = df_sub[v].apply(lambda x: x).max()
        max_neg_error = df_sub[v].apply(lambda x: x).min()
        text = r"\textit{%.2f}" % mae
        text += "\n"
        text += r"\textbf{%.2f}" % rmse
        text += "\n"
        text += r"\textrm{%.2f}" % max_pos_error
        text += "\n"
        text += r"\textrm{%.2f}" % max_neg_error
        annotations.append((cnt, m, text))
        cnt += 1

    pd.set_option("display.max_columns", None)
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    gs = gridspec.GridSpec(
        2, 1, height_ratios=[0.22, 1]
    )  # Adjust height ratios to change the size of subplots
    ax = plt.subplot(gs[1])  # This will create the subplot for the main violin plot.
    vplot = ax.violinplot(
        vData,
        showmeans=True,
        showmedians=False,
        showextrema=False,
        quantiles=[[0.05, 0.95] for i in range(len(vData))],
        widths=widths,
    )
    for n, partname in enumerate(["cmeans"]):
        vp = vplot[partname]
        vp.set_edgecolor("black")
        vp.set_linewidth(1)
        vp.set_alpha(1)
    quantile_color = "red"
    quantile_style = "-"
    quantile_linewidth = 0.8
    for n, partname in enumerate(["cquantiles"]):
        vp = vplot[partname]
        vp.set_edgecolor(quantile_color)
        vp.set_linewidth(quantile_linewidth)
        vp.set_linestyle(quantile_style)
        vp.set_alpha(1)

    colors = ["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))]
    for n, pc in enumerate(vplot["bodies"], 1):
        pc.set_facecolor(colors[n - 1])
        pc.set_alpha(0.6)

    vLabels.insert(0, "")
    xs = [i for i in range(len(vLabels))]
    xs_error = [i for i in range(-1, len(vLabels) + 1)]
    ax.plot(
        xs_error,
        [1 for i in range(len(xs_error))],
        "k--",
        label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        xs_error,
        [0 for i in range(len(xs_error))],
        "k--",
        linewidth=0.5,
        alpha=0.5,
        # label=r"Reference Energy",
        zorder=0,
    )
    ax.plot(
        xs_error,
        [-1 for i in range(len(xs_error))],
        "k--",
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        [],
        [],
        linestyle=quantile_style,
        color=quantile_color,
        linewidth=quantile_linewidth,
        label=r"5-95th Percentile",
    )
    navy_blue = (0.0, 0.32, 0.96)
    ax.set_xticks(xs)
    plt.setp(
        ax.set_xticklabels(vLabels),
        rotation=x_label_rotation,
        fontsize=x_label_fontsize,
    )
    ax.set_xlim((0, len(vLabels)))
    if ylim is not None:
        ax.set_ylim(ylim)
        major_yticks, minor_yticks = create_minor_y_ticks(ylim)
        ax.set_yticks(major_yticks)
        ax.set_yticks(minor_yticks, minor=True)

    lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

    if set_xlabel:
        ax.set_xlabel("Level of Theory", color="k")
    ax.set_ylabel(ylabel, color="k")

    ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
    for n, xtick in enumerate(ax.get_xticklabels()):
        xtick.set_color(colors[n - 1])
        xtick.set_alpha(0.8)

    ax_error = plt.subplot(gs[0], sharex=ax)
    # ax_error.spines['top'].set_visible(False)
    ax_error.spines["right"].set_visible(False)
    ax_error.spines["left"].set_visible(False)
    ax_error.spines["bottom"].set_visible(False)
    ax_error.tick_params(left=False, labelleft=False, bottom=False, labelbottom=False)

    # Synchronize the x-limits with the main subplot
    ax_error.set_xlim((0, len(vLabels)))
    ax_error.set_ylim(0, 1)  # Assuming the upper subplot should have no y range
    error_labels = r"\textit{MAE}"
    error_labels += "\n"
    error_labels += r"\textbf{RMSE}"
    error_labels += "\n"
    error_labels += r"\textrm{MaxE}"
    error_labels += "\n"
    error_labels += r"\textrm{MinE}"
    ax_error.annotate(
        error_labels,
        xy=(0, 1),  # Position at the vertical center of the narrow subplot
        xytext=(0, 0.2),
        color="black",
        fontsize=f"{table_fontsize}",
        ha="center",
        va="center",
    )
    for idx, (x, y, text) in enumerate(annotations):
        ax_error.annotate(
            text,
            xy=(x, 1),  # Position at the vertical center of the narrow subplot
            # xytext=(0, 0),
            xytext=(x, 0.2),
            color="black",
            fontsize=f"{table_fontsize}",
            ha="center",
            va="center",
        )

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
    path = f"{output_basename}_violin.{ext}"
    print(f"{path}")
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    plt.clf()
    return


def violin_plot_table_multi(
    dfs,
    df_labels_and_columns: {},
    output_filename: str,
    plt_title: str = None,
    bottom: float = 0.4,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlable=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    table_fontsize=8,
    ylabel=r"Error (kcal$\cdot$mol$^{-1}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
    grid_heights=None,
    grid_widths=None,
    mcure=None,
    error_labels_position=(0, 0.25),
    violin_alpha=0.6,
) -> None:
    """
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        Example:
        dfs = [
            {"df": df_jdz, "label": "jun-cc-pVDZ", "ylim": [-4, 2]},
            {"df": df_adz, "label": "aug-cc-pVDZ", "ylim": [-2, 2]},
            {"df": df_atz, "label": "aug-cc-pVTZ", "ylim": [-2, 2]},
            {"df": df_adtz, "label": "aug-cc-pV[DT]Z", "ylim": [-2, 2]},
        ]
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
        mcure: If requested, must pre-compute MCURE for each df_labels_and_columns key and assign as a dictionary
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import gridspec

    ylabel_initial = ylabel

    print(f"Plotting {output_filename}")
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    if grid_heights is None:
        grid_heights = []
        for i in range(len(dfs)):
            grid_heights.append(0.6)
            grid_heights.append(2)
    print(len(dfs) * 2)

    gs = gridspec.GridSpec(
        len(dfs) * 2, 1, height_ratios=grid_heights
    )  # Adjust height ratios to change the size of subplots
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    for ind_0, j in enumerate(dfs):
        df = j["df"]
        subplot_label = j["label"]
        ylim = j["ylim"]
        vLabels, vData = [], []
        annotations = []  # [(x, y, text), ...]
        cnt = 1
        ind = 2 * ind_0
        print(f"{ind = }, {subplot_label = }")
        plt.rcParams["text.usetex"] = usetex
        non_null = len(df)
        for k, v in df_labels_and_columns.items():
            df[v] = pd.to_numeric(df[v])
            df_sub = df[df[v].notna()].copy()
            local_value = df_sub[v].to_list()
            if len(local_value) == 0:
                local_value = [0] * len(vData[-1])
            vData.append(local_value)
            k_label = "\\textbf{" + k + "}"
            k_label = convert_deltas_ssapt0(k_label)
            vLabels.append(k_label)
            m = df_sub[v].max()
            rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
            mae = df_sub[v].apply(lambda x: abs(x)).mean()
            max_pos_error = df_sub[v].apply(lambda x: x).max()
            max_neg_error = df_sub[v].apply(lambda x: x).min()
            text = r"\textit{%.2f}" % mae
            text += "\n"
            text += r"\textbf{%.2f}" % rmse
            text += "\n"
            text += r"\textrm{%.2f}" % max_pos_error
            text += "\n"
            text += r"\textrm{%.2f}" % max_neg_error
            if mcure is not None:
                text += "\n"
                text += r"\textrm{%.2f}" % mcure[k][ind_0]
            annotations.append((cnt, m, text))
            cnt += 1
            tmp = df_sub[v].notna().sum()
            if tmp < non_null and tmp != 0:
                non_null = tmp

        pd.set_option("display.max_columns", None)
        ax = plt.subplot(
            gs[ind + 1]
        )  # This will create the subplot for the main violin plot.
        vplot = ax.violinplot(
            vData,
            showmeans=True,
            showmedians=False,
            showextrema=False,
            quantiles=[[0.05, 0.95] for i in range(len(vData))],
            widths=widths,
        )
        for n, partname in enumerate(["cmeans"]):
            vp = vplot[partname]
            vp.set_edgecolor("black")
            vp.set_linewidth(1)
            vp.set_alpha(1)
        quantile_color = "red"
        quantile_style = "-"
        quantile_linewidth = 0.8
        for n, partname in enumerate(["cquantiles"]):
            vp = vplot[partname]
            vp.set_edgecolor(quantile_color)
            vp.set_linewidth(quantile_linewidth)
            vp.set_linestyle(quantile_style)
            vp.set_alpha(1)

        if colors is None:
            colors = ["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))]
        for n, pc in enumerate(vplot["bodies"], 1):
            pc.set_facecolor(colors[n - 1])
            pc.set_alpha(violin_alpha)

        vLabels.insert(0, "")
        xs = [i for i in range(len(vLabels))]
        xs_error = [i for i in range(-1, len(vLabels) + 1)]
        ax.plot(
            xs_error,
            [1 for i in range(len(xs_error))],
            "k--",
            label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
            zorder=0,
            linewidth=0.6,
        )
        ax.plot(
            xs_error,
            [0 for i in range(len(xs_error))],
            "k--",
            linewidth=0.5,
            alpha=0.5,
            # label=r"Reference Energy",
            zorder=0,
        )
        ax.plot(
            xs_error,
            [-1 for i in range(len(xs_error))],
            "k--",
            zorder=0,
            linewidth=0.6,
        )
        ax.plot(
            [],
            [],
            linestyle=quantile_style,
            color=quantile_color,
            linewidth=quantile_linewidth,
            label=r"5-95th Percentile",
        )
        navy_blue = (0.0, 0.32, 0.96)
        ax.set_xticks(xs)
        plt.setp(
            ax.set_xticklabels(vLabels),
            rotation=x_label_rotation,
            fontsize=x_label_fontsize,
        )
        ax.set_xlim((0, len(vLabels)))
        if ylim is not None:
            ax.set_ylim(ylim)
            major_yticks, minor_yticks = create_minor_y_ticks(ylim)
            ax.set_yticks(major_yticks)
            ax.set_yticks(minor_yticks, minor=True)

        if ind == 0:
            lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

        if set_xlable:
            ax.set_xlabel("Level of Theory", color="k")
        # ax.set_ylabel(f"{subplot_label}\n{ylabel_initial}", color="k")
        ax.set_ylabel(f"{ylabel_initial}", color="k")

        ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
        for n, xtick in enumerate(ax.get_xticklabels()):
            xtick.set_color(colors[n - 1])
            # xtick.set_alpha(0.8)
            xtick.set_alpha(violin_alpha)

        if ind != len(dfs) * 2 - 2:
            # ax.spines["bottom"].set_visible(False)
            # ax.tick_params(bottom=False)
            ax.tick_params(
                left=True,
                labelleft=True,
                bottom=False,
                labelbottom=False,
            )

            # plt.setp(ax.xaxis.get_ticklabels(), visible=False)
            # do not have xlabels

        ax_error = plt.subplot(gs[ind], sharex=ax)
        # ax_error.spines['top'].set_visible(False)
        ax_error.spines["right"].set_visible(False)
        ax_error.spines["left"].set_visible(False)
        ax_error.spines["bottom"].set_visible(False)
        ax_error.tick_params(
            left=False, labelleft=False, bottom=False, labelbottom=False
        )

        # Synchronize the x-limits with the main subplot
        ax_error.set_xlim((0, len(vLabels)))
        ax_error.set_ylim(0, 1)  # Assuming the upper subplot should have no y range
        error_labels = r"\textit{MAE}"
        error_labels += "\n"
        error_labels += r"\textbf{RMSE}"
        error_labels += "\n"
        error_labels += r"\textrm{MaxE}"
        error_labels += "\n"
        error_labels += r"\textrm{MinE}"
        if mcure is not None:
            error_labels += "\n"
            error_labels += r"\textrm{MCURE}"

        subplot_title = r"\textbf{" + subplot_label + r"}"
        subplot_title += r"(\textbf{" + str(non_null) + r"})" 
        ax_error.set_title(subplot_title, pad=-4)

        ax_error.annotate(
            error_labels,
            xy=(0, 1),  # Position at the vertical center of the narrow subplot
            xytext=error_labels_position,
            color="black",
            fontsize=f"{table_fontsize}",
            ha="center",
            va="center",
        )
        for idx, (x, y, text) in enumerate(annotations):
            ax_error.annotate(
                text,
                xy=(x, 1),  # Position at the vertical center of the narrow subplot
                # xytext=(0, 0),
                xytext=(x, 0.25),
                color="black",
                fontsize=f"{table_fontsize}",
                ha="center",
                va="center",
            )

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
    path = f"{output_basename}_violin.{ext}"
    print(f"{path}")
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    plt.clf()
    return

def violin_plot_table_multi_horizontal(
    dfs,
    df_labels_and_columns: {},
    output_filename: str,
    plt_title: str = None,
    bottom: float = 0.4,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlable=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    table_fontsize=8,
    ylabel=r"Error (kcal$\cdot$mol$^{-1}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
    grid_heights=None,
    grid_widths=None,
    mcure=None,
) -> None:
    """
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        Example:
        dfs = [
            {"df": df_all, "label": "4569", "ylim": [-4, 2]},
            {"df": df_subset, "label": "288", "ylim": [-2, 2]},
        ]
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
        mcure: If requested, must pre-compute MCURE for each df_labels_and_columns key and assign as a dictionary
    """
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import gridspec

    ylabel_initial = ylabel

    print(f"Plotting {output_filename}")
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    if grid_heights is None:
        grid_heights = []
        for i in range(len(dfs)):
            grid_heights.append(0.6)
            grid_heights.append(2)
    print(len(dfs) * 2)

    gs = gridspec.GridSpec(
        2, len(dfs), height_ratios=grid_heights, width_ratios=grid_widths
    )  # Adjust height ratios to change the size of subplots
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    vlabels_df = []
    for ind_0, j in enumerate(dfs):
        df = j["df"]
        subplot_label = j["label"]
        ylim = j["ylim"]
        vLabels, vData = [], []
        annotations = []  # [(x, y, text), ...]
        cnt = 1
        ind = 2 * ind_0
        print(f"{ind = }, {subplot_label = }")
        plt.rcParams["text.usetex"] = usetex
        non_null = len(df)
        for k, v in df_labels_and_columns.items():
            if v not in df.columns:
                print(f"{v} not found in df. Skipping...")
                continue
            df[v] = pd.to_numeric(df[v])
            df_sub = df[df[v].notna()].copy()
            vData.append(df_sub[v].to_list())
            k_label = "\\textbf{" + k + "}"
            k_label = convert_deltas_ssapt0(k_label)
            vLabels.append(k_label)
            m = df_sub[v].max()
            rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
            mae = df_sub[v].apply(lambda x: abs(x)).mean()
            max_pos_error = df_sub[v].apply(lambda x: x).max()
            max_neg_error = df_sub[v].apply(lambda x: x).min()
            text = r"\textit{%.2f}" % mae
            text += "\n"
            text += r"\textbf{%.2f}" % rmse
            text += "\n"
            text += r"\textrm{%.2f}" % max_pos_error
            text += "\n"
            text += r"\textrm{%.2f}" % max_neg_error
            if mcure is not None:
                text += "\n"
                text += r"\textrm{%.2f}" % mcure[k][ind_0]
            annotations.append((cnt, m, text))
            cnt += 1
            tmp = df_sub[v].notna().sum()
            if tmp < non_null:
                non_null = tmp
        vlabels_df.append(vLabels)

        pd.set_option("display.max_columns", None)
        ax = plt.subplot(
            gs[ind_0 + len(dfs)]
        )  # This will create the subplot for the main violin plot.
        vplot = ax.violinplot(
            vData,
            showmeans=True,
            showmedians=False,
            showextrema=False,
            quantiles=[[0.05, 0.95] for i in range(len(vData))],
            widths=widths,
        )
        for n, partname in enumerate(["cmeans"]):
            vp = vplot[partname]
            vp.set_edgecolor("black")
            vp.set_linewidth(1)
            vp.set_alpha(1)
        quantile_color = "red"
        quantile_style = "-"
        quantile_linewidth = 0.8
        for n, partname in enumerate(["cquantiles"]):
            vp = vplot[partname]
            vp.set_edgecolor(quantile_color)
            vp.set_linewidth(quantile_linewidth)
            vp.set_linestyle(quantile_style)
            vp.set_alpha(1)

        colors = ["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))]
        for n, pc in enumerate(vplot["bodies"], 1):
            pc.set_facecolor(colors[n - 1])
            pc.set_alpha(0.6)

        vLabels.insert(0, "")
        xs = [i for i in range(len(vLabels))]
        xs_error = [i for i in range(-1, len(vLabels) + 1)]
        ax.plot(
            xs_error,
            [1 for i in range(len(xs_error))],
            "k--",
            label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
            zorder=0,
            linewidth=0.6,
        )
        ax.plot(
            xs_error,
            [0 for i in range(len(xs_error))],
            "k--",
            linewidth=0.5,
            alpha=0.5,
            # label=r"Reference Energy",
            zorder=0,
        )
        ax.plot(
            xs_error,
            [-1 for i in range(len(xs_error))],
            "k--",
            zorder=0,
            linewidth=0.6,
        )
        ax.plot(
            [],
            [],
            linestyle=quantile_style,
            color=quantile_color,
            linewidth=quantile_linewidth,
            label=r"5-95th Percentile",
        )
        navy_blue = (0.0, 0.32, 0.96)
        ax.set_xticks(xs)
        plt.setp(
            ax.set_xticklabels(vLabels),
            rotation=x_label_rotation,
            fontsize=x_label_fontsize,
        )
        ax.set_xlim((0, len(vLabels)))
        if ylim is not None:
            ax.set_ylim(ylim)
            major_yticks, minor_yticks = create_minor_y_ticks(ylim)
            ax.set_yticks(major_yticks)
            ax.set_yticks(minor_yticks, minor=True)

        if ind == 0:
            lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

        if set_xlable:
            ax.set_xlabel("Level of Theory", color="k")
        # ax.set_ylabel(f"{subplot_label}\n{ylabel_initial}", color="k")
        ax.set_ylabel(f"{ylabel_initial}", color="k")

        ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
        for n, xtick in enumerate(ax.get_xticklabels()):
            xtick.set_color(colors[n - 1])
            xtick.set_alpha(0.8)


        ax_error = plt.subplot(gs[ind_0], sharex=ax)
        # ax_error.spines['top'].set_visible(False)
        ax_error.spines["right"].set_visible(False)
        ax_error.spines["left"].set_visible(False)
        ax_error.spines["bottom"].set_visible(False)
        ax_error.tick_params(
            left=False, labelleft=False, bottom=False, labelbottom=False
        )

        # Synchronize the x-limits with the main subplot
        ax_error.set_xlim((0, len(vLabels)))
        ax_error.set_ylim(0, 1)  # Assuming the upper subplot should have no y range
        error_labels = r"\textit{MAE}"
        error_labels += "\n"
        error_labels += r"\textbf{RMSE}"
        error_labels += "\n"
        error_labels += r"\textrm{MaxE}"
        error_labels += "\n"
        error_labels += r"\textrm{MinE}"
        if mcure is not None:
            error_labels += "\n"
            error_labels += r"\textrm{MCURE}"

        subplot_title = r"\textbf{" + subplot_label + r"}"
        subplot_title += r"(\textbf{" + str(non_null) + r"})" 
        ax_error.set_title(subplot_title, pad=-4)

        ax_error.annotate(
            error_labels,
            xy=(0, 1),  # Position at the vertical center of the narrow subplot
            xytext=(0, 0.25),
            color="black",
            fontsize=f"{table_fontsize}",
            ha="center",
            va="center",
        )
        for idx, (x, y, text) in enumerate(annotations):
            ax_error.annotate(
                text,
                xy=(x, 1),  # Position at the vertical center of the narrow subplot
                # xytext=(0, 0),
                xytext=(x, 0.25),
                color="black",
                fontsize=f"{table_fontsize}",
                ha="center",
                va="center",
            )

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
    path = f"{output_basename}_violin.{ext}"
    print(f"{path}")
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    plt.clf()
    return

def convert_deltas_ssapt0(k_label):
    k_label = k_label.replace("DMP2", r"$\delta$MP2").replace("SSAPT0", r"\emph{s}SAPT0").replace("WB97", r"$\omega$B97")
    return k_label

def violin_plot_table_multi_SAPT_components(
    dfs,
    df_labels_and_columns_elst={},
    df_labels_and_columns_exch={},
    df_labels_and_columns_indu={},
    df_labels_and_columns_disp={},
    df_labels_and_columns_total={},
    output_filename: str = "output",
    plt_title: str = None,
    bottom: float = 0.4,
    left: float = None,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlabel=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    y_label_fontsize=8,
    table_fontsize=8,
    title_fontsize=10,
    ylabel=r"Error (kcal$\cdot$mol$^{-1}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
    grid_heights=None,
    grid_widths=None,
    MAE="textit",
    RMSE="textbf",
    MaxE="textrm",
    MinE="textrm",
    mcure=None,
    wspace=None,
    annotations_texty=0.4,
    share_y_axis=False,
    table_delimiter="",
    gridlines_linewidths=0.6,
    violin_alphas=0.6,
    quantile_color = "red",
    quantile_style = "-",
    quantile_linewidth = 0.8,
    pm_alpha=1.0,
    zero_alpha=0.5,
    hide_ytick_label_edges=False,
    add_title=True,
    ylabel_count=True,
    disable_xtick_labels=False,
    bbox_inches="tight",
) -> None:
    """
    TODO: maybe a 4xN grid for the 4 components of SAPT?
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
        mcure: If requested, must pre-compute MCURE for each components df_labels_and_columns_<COMPONENT>
        Dictionary structure:
            mcure_labels = {
                "ELST": [],
                "EXCH": [],
                "IND": [],
                "DISP": [],
            }
    """
    colors_initialized = True
    if colors is None:
        colors_initialized = False
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import gridspec

    ylabel_initial = ylabel

    print(f"Plotting {output_filename}")
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    if grid_heights is None:
        heights = []
        for i in range(len(dfs)):
            heights.append(0.4)
            heights.append(2)
    if grid_widths is None:
        grid_widths = [1, 1, 1, 2]
    print(f"Must have {len(dfs) * 2} heights")
    columns = 0
    sapt_terms_plot = [] 
    if len(df_labels_and_columns_elst) > 0:
        columns += 1
        sapt_terms_plot.append("ELST")
    if len(df_labels_and_columns_exch) > 0:
        columns += 1
        sapt_terms_plot.append("EXCH")
    if len(df_labels_and_columns_indu) > 0:
        columns += 1
        sapt_terms_plot .append("INDU")
    if len(df_labels_and_columns_disp) > 0:
        columns += 1
        sapt_terms_plot.append("DISP")
    if len(df_labels_and_columns_total) > 0:
        columns += 1
        sapt_terms_plot.append("TOTAL")


    gs = gridspec.GridSpec(
        len(dfs) * 2, columns, height_ratios=grid_heights, width_ratios=grid_widths,
    )  # Adjust height ratios to change the size of subplots
    ax1 = None
    axs = []
    if wspace is not None:
        gs.update(wspace=wspace)
    print(f"{gs = }")
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    for nn, term in enumerate(sapt_terms_plot):
        if term == "ELST":
            df_labels_and_columns = df_labels_and_columns_elst
            sapt_color = 'red'
        elif term == "EXCH":
            df_labels_and_columns = df_labels_and_columns_exch
            sapt_color = 'green'
        elif term == "INDU":
            df_labels_and_columns = df_labels_and_columns_indu
            sapt_color = 'blue'
        elif term == "DISP":
            df_labels_and_columns = df_labels_and_columns_disp
            sapt_color = 'orange'
        elif term == "TOTAL":
            df_labels_and_columns = df_labels_and_columns_total
            sapt_color = 'black'
        print(f"{term = }")
        for ind_0, j in enumerate(dfs):
            df = j["df"]
            subplot_label = j["label"]
            ylim = j["ylim"][nn]
            vLabels, vData = [], []
            annotations = []  # [(x, y, text), ...]
            cnt = 1
            ind = ind_0 * 2
            plt.rcParams["text.usetex"] = usetex
            non_null = len(df)
            # print(f"{j['basis']}, {non_null = }")
# -            vData.append(df_sub[v].to_list())
# +            local_value = df_sub[v].to_list()
# +            if len(local_value) == 0:
# +                local_value = [0] * len(vData[-1])
# +            vData.append(local_value)
#              k_label = "\\textbf{" + k + "}"
#              k_label = convert_deltas(k_label)
#              vLabels.append(k_label)
# @@ -549,7 +552,7 @@ def violin_plot_table_multi(
#              annotations.append((cnt, m, text))
#              cnt += 1
#              tmp = df_sub[v].notna().sum()
# -            if tmp < non_null:
# +            if tmp < non_null and tmp != 0:
#                  non_null = tmp
            for col_ind, (k, v) in enumerate(df_labels_and_columns.items()):
                df[v] = pd.to_numeric(df[v])
                df_sub = df[df[v].notna()].copy()
                if len(df_sub) != len(df):
                    print('Missing data in', k, v)
                local_value = df_sub[v].to_list()
                if len(local_value) == 0:
                    local_value = [0] * len(vData[-1])
                vData.append(local_value)
                k_label = r"\noindent\textbf{" + r"}\\\textbf{".join(k.split(r"\\")) + "}"
                # k_label = r"\noindent\textbf{" + k + "}"
                k_label = convert_deltas_ssapt0(k_label)
                vLabels.append(k_label)
                m = df_sub[v].max()
                rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
                mae = df_sub[v].apply(lambda x: abs(x)).mean()
                max_pos_error = df_sub[v].apply(lambda x: x).max()
                max_neg_error = df_sub[v].apply(lambda x: x).min()
                errors_ls = []
                l_delim = table_delimiter if col_ind != len(df_labels_and_columns.keys()) - 1 else ""
                if MAE:
                    errors_ls.append(rf"\{MAE}{{{mae:.2f}}}{l_delim}")
                if RMSE:
                    errors_ls.append(rf"\{RMSE}{{{rmse:.2f}}}{l_delim}")
                if MaxE:
                    errors_ls.append(rf"\{MaxE}{{{max_pos_error:.2f}}}{l_delim}")
                if MinE:
                    errors_ls.append(rf"\{MinE}{{{max_neg_error:.2f}}}{l_delim}")
                # if mcure is not None and term != "TOTAL":
                if mcure is not None:
                    # try:
                        errors_ls.append(rf"\textrm{mcure[term][k][ind_0]:.2f}{l_delim}")
                    # except (Exception) as e:
                    #     print(f"Error: {e}")
                    #     print(f"term: {term}, k: {k}, ind_0: {ind_0}")
                    #     import sys
                    #     sys.exit(1)
                text = "\n".join(errors_ls)
                annotations.append((cnt, m, text))
                cnt += 1
                tmp = df_sub[v].notna().sum()
                if tmp < non_null and tmp != 0:
                    non_null = tmp

            pd.set_option("display.max_columns", None)
            if share_y_axis and ind != 0:
                ax = plt.subplot(
                    gs[ind+1, nn],
                    sharey=ax1,
                )
            else:
                ax = plt.subplot(
                    gs[ind+1, nn]
                )
                ax1 = ax
            axs.append(ax)
            vplot = ax.violinplot(
                vData,
                showmeans=True,
                showmedians=False,
                showextrema=False,
                quantiles=[[0.05, 0.95] for i in range(len(vData))],
                widths=widths,
            )
            for n, partname in enumerate(["cmeans"]):
                vp = vplot[partname]
                vp.set_edgecolor("black")
                vp.set_linewidth(1)
                vp.set_alpha(1)
            for n, partname in enumerate(["cquantiles"]):
                vp = vplot[partname]
                vp.set_edgecolor(quantile_color)
                vp.set_linewidth(quantile_linewidth)
                vp.set_linestyle(quantile_style)
                vp.set_alpha(1)

            if not colors_initialized:
                colors = [["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))] for i in range(columns)]
            else:
                print(f"Term: {nn} colors {colors[nn]}")
            for n, pc in enumerate(vplot["bodies"], 1):
                try:
                    pc.set_facecolor(colors[nn][n - 1])
                except IndexError:
                    pc.set_facecolor('pink')
                pc.set_alpha(violin_alphas)

            vLabels.insert(0, "")
            xs = [i for i in range(len(vLabels))]
            xs_error = [i for i in range(-1, len(vLabels) + 1)]
            ax.plot(
                xs_error,
                [1 for i in range(len(xs_error))],
                "k--",
                label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
                zorder=0,
                alpha=pm_alpha,
                linewidth=gridlines_linewidths,
            )
            ax.plot(
                xs_error,
                [0 for i in range(len(xs_error))],
                "k--",
                linewidth=gridlines_linewidths,
                alpha=zero_alpha,
                # label=r"Reference Energy",
                zorder=0,
            )
            ax.plot(
                xs_error,
                [-1 for i in range(len(xs_error))],
                "k--",
                zorder=0,
                alpha=pm_alpha,
                linewidth=gridlines_linewidths,
            )
            ax.plot(
                [],
                [],
                linestyle=quantile_style,
                color=quantile_color,
                linewidth=quantile_linewidth,
                label=r"5-95th Percentile",
            )
            navy_blue = (0.0, 0.32, 0.96)
            ax.set_xticks(xs)
            ax.spines['top'].set_color(sapt_color)
            ax.spines["right"].set_color(sapt_color)
            ax.spines["left"].set_color(sapt_color)
            ax.spines["bottom"].set_color(sapt_color)
            ax.spines['top'].set_linewidth(2.5)
            ax.spines["right"].set_linewidth(2.5)
            ax.spines["left"].set_linewidth(2.5)
            ax.spines["bottom"].set_linewidth(2.5)
            plt.setp(
                ax.set_xticklabels(vLabels),
                rotation=x_label_rotation,
                fontsize=x_label_fontsize,
            )
            ax.set_xlim((0, len(vLabels)))
            if ylim is not None:
                ax.set_ylim(ylim)
                if not share_y_axis or nn == 0:
                    print(ind, nn, 'ylim')
                    major_yticks, minor_yticks = create_minor_y_ticks(ylim)
                    ax.set_yticks(major_yticks)
                    ax.set_yticks(minor_yticks, minor=True)
                    if hide_ytick_label_edges:
                        major_yticks = list(major_yticks)
                        major_yticks[0] = ""
                        major_yticks[-1] = ""

                    ax.set_yticklabels(
                        major_yticks,
                        fontsize=y_label_fontsize,
                    )
                else:
                    ax.set_yticks([])

            if ind == 0 and nn == columns - 1 and legend_loc is not None:
                lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

            if set_xlabel:
                ax.set_xlabel("Level of Theory", color="k")
            # ax.set_ylabel(f"{subplot_label}\n{ylabel_initial}", color="k")
            if nn == 0:
                ylabel_row = r"\textbf{" + subplot_label + r"}"
                if ylabel_count:
                    ylabel_row += r"(\textbf{" + str(non_null) + r"})" 
                if ylabel_initial is not None:
                    ylabel_row += f"\n{ylabel_initial}"
                ax.set_ylabel(ylabel_row, color="k", fontsize=y_label_fontsize)

            ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
            # ax.grid(color="#54585A", which="minor", linewidth=0.5, alpha=0.5)

            for n, xtick in enumerate(ax.get_xticklabels()):
                try:
                    xtick.set_color(colors[nn][n - 1])
                except IndexError:
                    xtick.set_color('pink')
                xtick.set_alpha(0.8)

            if ind != len(dfs) * 2 - 2:
                ax.tick_params(
                    left=True,
                    labelleft=True,
                    bottom=False,
                    labelbottom=False,
                )

            if disable_xtick_labels:
                plt.setp(ax.xaxis.get_ticklabels(), visible=False)
                # ax.spines["bottom"].set_visible(False)
                # ax.tick_params(bottom=False)
            ax_error = plt.subplot(gs[ind, nn], sharex=ax)
            ax_error.spines['top'].set_visible(False)
            ax_error.spines["right"].set_visible(False)
            ax_error.spines["left"].set_visible(False)
            ax_error.spines["bottom"].set_visible(False)
            ax_error.tick_params(
                left=False, labelleft=False, bottom=False, labelbottom=False
            )

            # Synchronize the x-limits with the main subplot
            if share_y_axis:
                ax.set_xlim((0.5, len(vLabels) - 0.5))
            else:
                ax.set_xlim((0, len(vLabels)))
            ax_error.set_ylim(0, 1)  # Assuming the upper subplot should have no y range
            error_labels = []

            if MAE:
                error_labels.append(rf"\{MAE}{{MAE}}")
            if RMSE:
                error_labels.append(rf"\{RMSE}{{RMSE}}")
            if MaxE:
                error_labels.append(rf"\{MaxE}{{MaxE}}")
            if MinE:
                error_labels.append(rf"\{MinE}{{MinE}}")
            if mcure is not None:
                error_labels.append(rf"\textrm{{MCURE}}")
            error_labels = "\n".join(error_labels)

            if ind == 0:
                if add_title:
                    ax_error.spines['top'].set_visible(True)
                    subplot_title = r"\textbf{" + str(term) + r"}"
                    ax_error.set_title(subplot_title, color=sapt_color, pad=-4, fontsize=title_fontsize)

            if not share_y_axis or nn == 0:
                x_pos = 0
                x_pos_shift = 0
                if share_y_axis:
                    x_pos = 0.5
                    x_pos_shift = -0.5
                ax_error.annotate(
                    error_labels,
                    xy=(x_pos, 1),  # Position at the vertical center of the narrow subplot
                    xytext=(x_pos + x_pos_shift, annotations_texty),
                    color="black",
                    fontsize=f"{table_fontsize}",
                    ha="right",
                    va="center",
                )
            for idx, (x, y, text) in enumerate(annotations):
                ax_error.annotate(
                    text,
                    xy=(x, 1),  # Position at the vertical center of the narrow subplot
                    # xytext=(0, 0),
                    xytext=(x, annotations_texty),
                    color="black",
                    fontsize=f"{table_fontsize}",
                    ha="center",
                    va="center",
                )

    if plt_title is not None:
        plt.title(f"{plt_title}")
        
    fig.subplots_adjust(bottom=bottom)
    if left:
        fig.subplots_adjust(left=left)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
    path = f"{output_basename}_violin.{ext}"
    print(f"{path}")
    if share_y_axis:
        # plt.subplots_adjust(hspace=0.1)
        plt.subplots_adjust(wspace=0.05)
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches=bbox_inches,
        dpi=dpi,
    )
    plt.clf()
    return


def violin_plot_table_multi_general(
    dfs,
    df_labels_and_columns={},
    output_filename: str = "output",
    plt_title: str = None,
    bottom: float = 0.4,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlable=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    y_label_fontsize=8,
    table_fontsize=8,
    title_fontsize=10,
    ylabel=r"Error (kcal$\cdot$mol$^{-1}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
    legend_loc="upper right",
    grid_heights=None,
    grid_widths=None,
    MA="textit",
    RMS="textbf",
    Max="textrm",
    Min="textrm",
    mcure=None,
    display_counts=None,
    wspace=None,
    annotations_texty=0.4,
    share_y_axis=False,
    table_delimiter="",
    gridlines_linewidths=0.6,
    violin_alphas=0.6,
    quantile_color = "red",
    quantile_style = "-",
    quantile_linewidth = 0.8,
    pm_alpha=1.0,
    zero_alpha=0.5,
    hide_ytick_label_edges=False,
) -> None:
    """
    TODO: maybe a 4xN grid for the 4 components of SAPT?
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
        mcure: If requested, must pre-compute MCURE for each components df_labels_and_columns_<COMPONENT>
        Dictionary structure:
            mcure_labels = {
                "ELST": [],
                "EXCH": [],
                "IND": [],
                "DISP": [],
            }
    """
    colors_initialized = True
    if colors is None:
        colors_initialized = False
    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import gridspec

    ylabel_initial = ylabel

    print(f"Plotting {output_filename}")
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    if grid_heights is None:
        heights = []
        for i in range(len(dfs)):
            heights.append(0.4)
            heights.append(2)
    if grid_widths is None:
        grid_widths = [1, 1, 1, 2]
    print(f"Must have {len(dfs) * 2} heights")
    columns = 0
    terms_plot = []
    if len(df_labels_and_columns) > 0:
        columns += 1
        terms_plot.append("Values")


    gs = gridspec.GridSpec(
        len(dfs) * 2, columns, height_ratios=grid_heights, width_ratios=grid_widths,
    )  # Adjust height ratios to change the size of subplots
    ax1 = None
    if wspace is not None:
        gs.update(wspace=wspace)
    print(f"{gs = }")
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    for nn, term in enumerate(terms_plot):
        term_color = 'black'
        print(f"{term = }")
        for ind_0, j in enumerate(dfs):
            df = j["df"]
            subplot_label = j["label"]
            ylim = j["ylim"][nn]
            vLabels, vData = [], []
            annotations = []  # [(x, y, text), ...]
            cnt = 1
            ind = ind_0 * 2
            plt.rcParams["text.usetex"] = usetex
            non_null = len(df)
            # print(f"{j['basis']}, {non_null = }")
            for col_ind, (k, v) in enumerate(df_labels_and_columns.items()):
                df[v] = pd.to_numeric(df[v])
                df_sub = df[df[v].notna()].copy()
                if len(df_sub) != len(df):
                    print('Missing data in', k, v)
                local_value = df_sub[v].to_list()
                # if len(local_value) == 0:
                #     local_value = [0] * len(vData[-1])
                vData.append(local_value)
                k_label = r"\noindent\textbf{" + r"}\\\textbf{".join(k.split(r"\\")) + "}"
                # k_label = r"\noindent\textbf{" + k + "}"
                k_label = convert_deltas_ssapt0(k_label)
                vLabels.append(k_label)
                m = df_sub[v].max()
                local_v = df_sub[v].dropna().to_numpy()
                rmse = np.sqrt(np.mean(local_v ** 2))
                mae = np.mean(np.abs(local_v))
                max_pos_error = local_v.max()
                max_neg_error = local_v.min()
                print(f"{k}, {len(local_v)}, {mae = :.2f}, {rmse = :.2f}, {max_pos_error = :.2f}, {max_neg_error = :.2f}")
                errors_ls = []
                l_delim = table_delimiter if col_ind != len(df_labels_and_columns.keys()) - 1 else ""
                if MA:
                    errors_ls.append(rf"\{MA}{{{mae:.2f}}}{l_delim}")
                if RMS:
                    errors_ls.append(rf"\{RMS}{{{rmse:.2f}}}{l_delim}")
                if Max:
                    errors_ls.append(rf"\{Max}{{{max_pos_error:.2f}}}{l_delim}")
                if Min:
                    errors_ls.append(rf"\{Min}{{{max_neg_error:.2f}}}{l_delim}")
                if mcure is not None and term != "TOTAL":
                    try:
                        errors_ls.append(rf"\textrm{mcure[term][k][ind_0]:.2f}{l_delim}")
                    except (Exception) as e:
                        print(f"Error: {e}")
                        print(f"term: {term}, k: {k}, ind_0: {ind_0}")
                        import sys
                        sys.exit(1)
                if display_counts is not None:
                    errors_ls.append(rf"\{Min}{{{len(local_v)}}}{l_delim}")
                text = "\n".join(errors_ls)
                annotations.append((cnt, m, text))
                cnt += 1
            pd.set_option("display.max_columns", None)
            if share_y_axis and ind != 0:
                ax = plt.subplot(
                    gs[ind+1, nn],
                    sharey=ax1,
                )
            else:
                ax = plt.subplot(
                    gs[ind+1, nn]
                )
                ax1 = ax
            vplot = ax.violinplot(
                vData,
                showmeans=True,
                showmedians=False,
                showextrema=False,
                quantiles=[[0.05, 0.95] for i in range(len(vData))],
                widths=widths,
            )
            for n, partname in enumerate(["cmeans"]):
                vp = vplot[partname]
                vp.set_edgecolor("black")
                vp.set_linewidth(1)
                vp.set_alpha(1)
            for n, partname in enumerate(["cquantiles"]):
                vp = vplot[partname]
                vp.set_edgecolor(quantile_color)
                vp.set_linewidth(quantile_linewidth)
                vp.set_linestyle(quantile_style)
                vp.set_alpha(1)

            if not colors_initialized:
                colors = [["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))] for i in range(columns)]
            else:
                print(f"Term: {nn} colors {colors[nn]}")
            for n, pc in enumerate(vplot["bodies"], 1):
                pc.set_facecolor(colors[nn][n - 1])
                pc.set_alpha(violin_alphas)

            vLabels.insert(0, "")
            xs = [i for i in range(len(vLabels))]
            xs_error = [i for i in range(-1, len(vLabels) + 1)]
            ax.plot(
                xs_error,
                [1 for i in range(len(xs_error))],
                "k--",
                label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
                zorder=0,
                alpha=pm_alpha,
                linewidth=gridlines_linewidths,
            )
            ax.plot(
                xs_error,
                [0 for i in range(len(xs_error))],
                "k--",
                linewidth=gridlines_linewidths,
                alpha=zero_alpha,
                # label=r"Reference Energy",
                zorder=0,
            )
            ax.plot(
                xs_error,
                [-1 for i in range(len(xs_error))],
                "k--",
                zorder=0,
                alpha=pm_alpha,
                linewidth=gridlines_linewidths,
            )
            ax.plot(
                [],
                [],
                linestyle=quantile_style,
                color=quantile_color,
                linewidth=quantile_linewidth,
                label=r"5-95th Percentile",
            )
            navy_blue = (0.0, 0.32, 0.96)
            ax.set_xticks(xs)
            ax.spines['top'].set_color(term_color)
            ax.spines["right"].set_color(term_color)
            ax.spines["left"].set_color(term_color)
            ax.spines["bottom"].set_color(term_color)
            ax.spines['top'].set_linewidth(2.5)
            ax.spines["right"].set_linewidth(2.5)
            ax.spines["left"].set_linewidth(2.5)
            ax.spines["bottom"].set_linewidth(2.5)
            plt.setp(
                ax.set_xticklabels(vLabels),
                rotation=x_label_rotation,
                fontsize=x_label_fontsize,
            )
            if share_y_axis:
                ax.set_xlim((0.5, len(vLabels) - 0.5))
            else:
                ax.set_xlim((0, len(vLabels)))
            if ylim is not None:
                ax.set_ylim(ylim)
                if not share_y_axis or nn == 0:
                    print(ind, nn, 'ylim')
                    major_yticks, minor_yticks = create_minor_y_ticks(ylim)
                    ax.set_yticks(major_yticks)
                    ax.set_yticks(minor_yticks, minor=True)
                    if hide_ytick_label_edges:
                        major_yticks = list(major_yticks)
                        major_yticks[0] = ""
                        major_yticks[-1] = ""

                    ax.set_yticklabels(
                        major_yticks,
                        fontsize=y_label_fontsize,
                    )
                else:
                    ax.set_yticks([])

            if ind == 0 and nn == columns - 1 and legend_loc is not None:
                lg = ax.legend(loc=legend_loc, edgecolor="black", fontsize="8")

            if set_xlable:
                ax.set_xlabel("Level of Theory", color="k")
            # ax.set_ylabel(f"{subplot_label}\n{ylabel_initial}", color="k")
            if nn == 0:
                ylabel_row = r"\textbf{" + subplot_label + r"}"
                ylabel_row += r"(\textbf{" + str(non_null) + r"})" f"\n{ylabel_initial}"
                ax.set_ylabel(ylabel_row, color="k", fontsize=y_label_fontsize)

            ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
            # ax.grid(color="#54585A", which="minor", linewidth=0.5, alpha=0.5)
            for n, xtick in enumerate(ax.get_xticklabels()):
                xtick.set_color(colors[nn][n - 1])
                xtick.set_alpha(0.8)

            if ind != len(dfs) * 2 - 2:
                ax.tick_params(
                    left=True,
                    labelleft=True,
                    bottom=False,
                    labelbottom=False,
                )

            ax_error = plt.subplot(gs[ind, nn], sharex=ax)
            ax_error.spines['top'].set_visible(False)
            ax_error.spines["right"].set_visible(False)
            ax_error.spines["left"].set_visible(False)
            ax_error.spines["bottom"].set_visible(False)
            ax_error.tick_params(
                left=False, labelleft=False, bottom=False, labelbottom=False
            )

            # Synchronize the x-limits with the main subplot
            ax_error.set_xlim((0, len(vLabels)))
            ax_error.set_ylim(0, 1)  # Assuming the upper subplot should have no y range
            error_labels = []

            if MA:
                error_labels.append(rf"\{MA}{{MA}}")
            if RMS:
                error_labels.append(rf"\{RMS}{{RMS}}")
            if Max:
                error_labels.append(rf"\{Max}{{Max}}")
            if Min:
                error_labels.append(rf"\{Min}{{Min}}")
            if mcure is not None:
                error_labels.append(rf"\textrm{{MCURE}}")
            if display_counts is not None:
                error_labels.append(rf"\{display_counts}{{Count}}")
            error_labels = "\n".join(error_labels)

            if ind == 0:
                ax_error.spines['top'].set_visible(True)
                # subplot_title = r"\textbf{" + str(term) + r"}" 
                # ax_error.set_title(subplot_title, color=term_color, pad=-4, fontsize=title_fontsize)

            if not share_y_axis or nn == 0:
                ax_error.annotate(
                    error_labels,
                    xy=(0, 1),  # Position at the vertical center of the narrow subplot
                    xytext=(0.0, annotations_texty),
                    color="black",
                    fontsize=f"{table_fontsize}",
                    ha="right",
                    va="center",
                )
            for idx, (x, y, text) in enumerate(annotations):
                ax_error.annotate(
                    text,
                    xy=(x, 1),  # Position at the vertical center of the narrow subplot
                    # xytext=(0, 0),
                    xytext=(x, annotations_texty),
                    color="black",
                    fontsize=f"{table_fontsize}",
                    ha="center",
                    va="center",
                )

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        output_basename, ext = (
            ".".join(output_filename.split(".")[:-1]),
            output_filename.split(".")[-1],
        )
    path = f"{output_basename}_violin.{ext}"
    print(f"{path}")
    plt.savefig(
        path,
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    plt.clf()
    return

if __name__ == "__main__":
    # Fake data generated for example
    import pandas as pd
    import numpy as np
    df = pd.DataFrame(
        {
            "MP2": 5 * np.random.randn(1000) + 0.5,
            "HF": 5 * np.random.randn(1000) - 0.5,
            "MP2.5": 5 * np.random.randn(1000) + 0.5,
        }
    )
    # Only specify columns you want to plot
    vals = {
        "MP2 label": "MP2",
        "HF label": "HF",
    }
    violin_plot(df, vals, ylim=[-20, 35], output_filename="example")



#
# @BEGIN LICENSE
#
# Psi4: an open-source quantum chemistry software package
#
# Copyright (c) 2007-2017 The Psi4 Developers.
#
# The copyrights for code used from other parties are included in
# the corresponding files.
#
# This file is part of Psi4.
#
# Psi4 is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3.
#
# Psi4 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Psi4; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# @END LICENSE
#

from __future__ import absolute_import
from __future__ import print_function
import re
import sys
import itertools
import collections

from cdsg_plot.modelchems import Method, BasisSet, Error, methods, bases, errors

try:
    dict.items
except AttributeError:
    # Python 3
    def itervalues(d):
        return iter(d.values())

    def items(d):
        return iter(d.items())

else:
    # Python 2
    def itervalues(d):
        return d.itervalues()

    def items(d):
        return d.items()


mc_archive = {"mtd": methods, "bas": bases, "err": errors}
fancy_mc_archive = {}
for tier in [methods, bases, errors]:
    for k, v in items(tier):
        fancy_mc_archive[k] = v.latex


# define helper functions for formatting table cells
def val(kw):
    return r"""%s%s""" % (kw["matelem"], kw["footnote"])


def graphics(kw):
    mc = "-".join([kw[bit] for bit in ["mtd", "opt", "bas"]])
    return r"""\includegraphics[width=6.67cm,height=3.5mm]{%s%s.pdf}""" % (
        kw["plotpath"],
        mc,
    )


def flat(kw):
    if kw["matelem"].strip():
        dbssmc = "-".join([kw[bit] for bit in ["dbse", "sset", "mtd", "opt", "bas"]])
        return r"""\includegraphics[width=6.67cm,height=3.5mm]{%sflat_%s.pdf}""" % (
            kw["plotpath"],
            dbssmc,
        )
    else:
        return ""


def liliowa(kw):
    if kw["matelem"].strip():
        dbssmc = "-".join([kw[bit] for bit in ["dbse", "sset", "mtd", "opt", "bas"]])
        return r"""\includegraphics[height=3.5mm]{%sliliowa_%s.pdf}""" % (
            kw["plotpath"],
            dbssmc,
        )
    else:
        return ""


def lmtdbas(kw):
    return """%-25s""" % (methods[kw["mtd"]].latex + "/" + bases[kw["bas"]].latex)


def label(kw):
    kwt = kw["target"]
    return """ %-25s""" % (
        mc_archive[kwt][kw[kwt]].latex if kwt in mc_archive else kw[kwt]
    )


def label2(kw):
    """This and fancy_mc_archive are experimental alternative for
    summoning up col/row headers. Safe so long as mtd/bas/opt have
    orthogonal keys.

    """
    try:
        return fancy_mc_archive[kw]
    except KeyError as e:
        print("""Consider adding {} to modelchems.py""".format(e))
        return kw


def count(kw):
    return r"""%6d""" % (kw["count"])


def empty(kw):
    return ""


def table_generic(
    dbse,
    serrors,
    mtd,
    bas,
    columnplan,
    rowplan=["bas", "mtd"],
    opt=["CP"],
    err=["mae"],
    sset=["default"],
    landscape=False,
    standalone=True,
    subjoin=True,
    suppressblanks=False,
    footnotes=[],
    title="",
    indextitle="",
    plotpath="",
    theme="",
):
    """
    Arrays *mtd* and *bas* contain the keys to the qcdb.Method and
    qcdb.BasisSet objects that span all those that the table may
    encompass. If method and basis are to be scanned over, the arrays
    should be in the desired order.

    """

    def table_header(kw, abbr, head1, head0, head2):
        """Form table header"""
        ref = r"""tbl:qcdb-%s-%s""" % (theme, "-".join([kw[bit] for bit in tag]))
        fancy_kw = {
            k: (mc_archive[k][v].latex if k in mc_archive else v) for k, v in items(kw)
        }
        text.append("")
        text.append(r"""\begingroup""")
        text.append(r"""\squeezetable""")
        text.append(
            r"""\begin{%s}[h!tp]""" % ("sidewaystable" if landscape else "table")
        )
        text.append(r"""\renewcommand{\baselinestretch}{1}""")
        text.append(r"""\caption{%s""" % (title.format(**fancy_kw).replace("_", "\\_")))
        text.append(r"""\label{%s}}""" % (ref))
        indices.append(
            r"""\scriptsize \ref{%s} & \scriptsize %s \\ """
            % (ref, indextitle.format(**fancy_kw))
        )
        text.append(r"""\begin{ruledtabular}""")
        text.append(r"""\begin{tabular}{%s}""" % (abbr))
        text.append(head1)
        text.append(head0)
        text.append(head2)
        text.append(hline)

    def table_footer():
        """Form table footer"""

        # search-and-replace footnotes
        fnmatch = re.compile(
            r"""(?P<cellpre>.*)"""
            + r"""(\\footnotemark)\{(?P<fntext>.*)\}"""
            + r"""(?P<cellpost>.*)"""
        )
        otfcounter = len(footnotes) + 1
        lines2replace = {}
        otffootnotes = collections.OrderedDict()
        for idx, line in enumerate(text):
            newcells = []
            changed = False
            for cell in line.split("&"):
                res = fnmatch.match(cell)
                if res:
                    if res.group("fntext") in otffootnotes:
                        localcounter = otffootnotes[res.group("fntext")]
                    else:
                        localcounter = otfcounter
                        if localcounter == 52:  # fn symbols run out at "zz"
                            otffootnotes["Missing some reactions"] = localcounter
                        else:
                            otfcounter += 1
                            otffootnotes[res.group("fntext")] = localcounter
                    newcells.append(
                        res.group("cellpre")
                        + r"""\footnotemark["""
                        + str(localcounter)
                        + "]"
                        + res.group("cellpost")
                    )
                    changed = True
                else:
                    newcells.append(cell)
            if changed:
                lines2replace[idx] = "&".join(newcells)
        for idx, line in items(lines2replace):
            text[idx] = line

        # search-and-suppress "blank" lines
        if suppressblanks:
            for idx, line in enumerate(text):
                if line.strip().endswith("\\"):
                    innards = "".join(line.rstrip(""" \\""").split("&")[1:])
                    if innards.isspace():
                        text[idx] = "%" + text[idx]

        # finish out table
        text.append(r"""\end{tabular}""")
        text.append(r"""\end{ruledtabular}""")
        for idx, fn in enumerate(footnotes):
            text.append(r"""\footnotetext[%d]{%s}""" % (idx + 1, fn))
        for fn, idx in items(otffootnotes):
            text.append(r"""\footnotetext[%d]{%s}""" % (idx, fn))
        text.append(r"""\end{%s}""" % ("sidewaystable" if landscape else "table"))
        text.append(r"""\endgroup""")
        text.append(r"""\clearpage""")
        text.append("")

    def matelem(dict_row, dict_col):
        """Return merge of index dictionaries *dict_row* and *dict_col* (precedence) with error string from serrors appended at key 'matelem'."""
        kw = dict(dict_row, **dict_col)
        errpiece = serrors["-".join([kw[bit] for bit in ["mtd", "opt", "bas"]])][
            kw["sset"]
        ][kw["dbse"]]
        kw["matelem"] = errpiece[kw["err"]]
        if "tgtcnt" in errpiece:
            kw["count"] = errpiece["tgtcnt"]
        if "misscnt" in errpiece and errpiece["misscnt"] != 0 and "tgtcnt" in errpiece:
            kw["footnote"] = r"""\footnotemark{Missing %d of %d reactions.}""" % (
                errpiece["misscnt"],
                errpiece["tgtcnt"],
            )
        else:
            kw["footnote"] = ""
        return kw

    # avoid misunderstandings
    keysincolumnplan = set(sum([list(col[-1].keys()) for col in columnplan], []))
    for key in ["dbse", "sset", "mtd", "opt", "bas", "err"]:
        if len(locals()[key]) > 1:
            if key not in rowplan and key not in keysincolumnplan:
                print(
                    """Warning: non-first values in argument '{0}' won't """
                    """get used. Add '{0}' to rowplan to iterate over """
                    """the values or add to columnplan to access"""
                    """different values.""".format(key)
                )
                sys.exit()

    # form LaTeX reference tag
    tag = []
    for key in ["dbse", "sset", "mtd", "opt", "bas", "err"]:
        if len(locals()[key]) == 1 or (key == rowplan[0] and not subjoin):
            tag.append(key)
    tag = set(tag)
    for col in columnplan:
        tag -= set(col[4].keys())

    # form column headers
    start = 1
    stop = 1
    head0 = ""
    for index in range(2, len(columnplan)):
        if columnplan[index][1] == columnplan[index - 1][1]:
            stop = index
        else:
            head0 += r"""\cline{%d-%d}""" % (start + 1, stop + 1)
            start = index
            stop = index
        if index + 1 == len(columnplan):
            head0 += r"""\cline{%d-%d}""" % (start + 1, stop + 1)

    abbr = "".join([col[0] for col in columnplan])
    h1 = [
        (k, len(list(g))) for k, g in itertools.groupby([col[1] for col in columnplan])
    ]
    head1 = (
        " & ".join(
            [
                r"""\multicolumn{%d}{c}{\textbf{%s}}""" % (repeat, label)
                for (label, repeat) in h1
            ]
        )
        + r""" \\ """
    )
    h2 = [
        (k, len(list(g))) for k, g in itertools.groupby([col[2] for col in columnplan])
    ]
    head2 = (
        " & ".join(
            [
                r"""\multicolumn{%d}{c}{\textbf{%s}}""" % (repeat, label)
                for (label, repeat) in h2
            ]
        )
        + r""" \\ """
    )

    # form table body
    text = []
    indices = []
    nH = len(rowplan)
    hline = r"""\hline"""
    kw = {
        "plotpath": plotpath,
        "sset": sset[0],
        "dbse": dbse[0],
        "err": err[0],
        "mtd": mtd[0],
        "opt": opt[0],
        "bas": bas[0],
    }

    if standalone:
        text += begin_latex_document()

    if nH == 1:
        subjoin = True

    if subjoin:
        table_header(kw, abbr, head1, head0, head2)
        if text[-1] != hline:
            text.append(hline)

    for hier0 in locals()[rowplan[0]]:
        kw[rowplan[0]] = hier0
        kw["target"] = rowplan[0]
        if nH > 1:

            if not subjoin:
                table_header(kw, abbr, head1, head0, head2)
            if text[-1] != hline:
                text.append(hline)
            # text.append(r"""\textbf{%s} \\ """ % (mc_archive[rowplan[0]][hier0].latex))
            text.append(r"""\textbf{%s} \\ """ % (label2(hier0)))

            for hier1 in locals()[rowplan[1]]:
                kw[rowplan[1]] = hier1
                kw["target"] = rowplan[1]
                if nH > 2:
                    # text.append(r"""\enspace\textbf{%s} \\ """ % (mc_archive[rowplan[1]][hier1].latex))
                    text.append(r"""\enspace\textbf{%s} \\ """ % (label2(hier1)))

                    for hier2 in locals()[rowplan[2]]:
                        kw[rowplan[2]] = hier2
                        kw["target"] = rowplan[2]

                        text.append(
                            r"""\enspace\enspace"""
                            + " & ".join(
                                [col[3](matelem(kw, col[4])) for col in columnplan]
                            )
                            + r""" \\ """
                        )
                else:
                    text.append(
                        r"""\enspace"""
                        + " & ".join(
                            [col[3](matelem(kw, col[4])) for col in columnplan]
                        )
                        + r""" \\ """
                    )
            if not subjoin:
                table_footer()
        else:
            text.append(
                " & ".join([col[3](matelem(kw, col[4])) for col in columnplan])
                + r""" \\ """
            )

    if subjoin:
        table_footer()

    if standalone:
        text += end_latex_document()

    return text, indices


def begin_latex_document():
    """Returns array of lines at start of LaTeX file that
    enable it to be compiled as its own document (standalone).

    """
    text = []
    text.append("")
    text.append(
        r"""\documentclass[aip,jcp,preprint,superscriptaddress,floatfix]{revtex4-1}"""
    )
    text.append(r"""\usepackage{bm}""")
    text.append(r"""\usepackage{dcolumn}""")
    text.append(r"""\usepackage{rotating}""")
    text.append(r"""\usepackage{longtable}""")
    text.append(r"""\usepackage[table]{xcolor}""")
    text.append(r"""\begin{document}""")
    text.append("")
    return text


def end_latex_document():
    """Returns array of lines at end of LaTeX file that
    enable it to be compiled as its own document (standalone).

    """
    text = []
    text.append("")
    text.append(r"""\end{document}""")
    return text


if __name__ == "__main__":
    from collections import OrderedDict

    serrors = {
        "MP2-CP-adtz": {
            "hb": {
                "S22": OrderedDict([("mae", "    0.38"), ("mape", "     2.5")]),
                "HBC1": OrderedDict([("mae", "    0.28"), ("mape", "     2.5")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    0.28"), ("mape", "     1.9")]),
                "DB4": OrderedDict([("mae", "    0.31"), ("mape", "     2.3")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    0.89"), ("mape", "    19.2")]),
                "HBC1": OrderedDict([("mae", "    0.28"), ("mape", "     2.5")]),
                "NBC1": OrderedDict([("mae", "    1.10"), ("mape", "   139.0")]),
                "HSG": OrderedDict([("mae", "    0.18"), ("mape", "     9.6")]),
                "DB4": OrderedDict([("mae", "    0.61"), ("mape", "    42.6")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    1.13"), ("mape", "    27.1")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    1.10"), ("mape", "   139.0")]),
                "HSG": OrderedDict([("mae", "    0.16"), ("mape", "    10.9")]),
                "DB4": OrderedDict([("mae", "    0.80"), ("mape", "    59.0")]),
            },
        },
        "CCSD-CP-adz": {
            "hb": {
                "S22": OrderedDict([("mae", "    2.42"), ("mape", "    18.0")]),
                "HBC1": OrderedDict([("mae", "    2.09"), ("mape", "    15.3")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    2.29"), ("mape", "    14.8")]),
                "DB4": OrderedDict([("mae", "    2.27"), ("mape", "    16.1")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    1.82"), ("mape", "    32.3")]),
                "HBC1": OrderedDict([("mae", "    2.09"), ("mape", "    15.3")]),
                "NBC1": OrderedDict([("mae", "    1.15"), ("mape", "   147.4")]),
                "HSG": OrderedDict([("mae", "    1.14"), ("mape", "    85.9")]),
                "DB4": OrderedDict([("mae", "    1.55"), ("mape", "    70.2")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    1.54"), ("mape", "    39.0")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    1.15"), ("mape", "   147.4")]),
                "HSG": OrderedDict([("mae", "    0.95"), ("mape", "    97.7")]),
                "DB4": OrderedDict([("mae", "    1.21"), ("mape", "    94.7")]),
            },
        },
        "MP2-CP-atz": {
            "hb": {
                "S22": OrderedDict([("mae", "    0.70"), ("mape", "     5.0")]),
                "HBC1": OrderedDict([("mae", "    0.52"), ("mape", "     4.1")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    0.61"), ("mape", "     4.0")]),
                "DB4": OrderedDict([("mae", "    0.61"), ("mape", "     4.4")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    0.86"), ("mape", "    17.0")]),
                "HBC1": OrderedDict([("mae", "    0.52"), ("mape", "     4.1")]),
                "NBC1": OrderedDict([("mae", "    0.98"), ("mape", "   126.7")]),
                "HSG": OrderedDict([("mae", "    0.24"), ("mape", "    13.5")]),
                "DB4": OrderedDict([("mae", "    0.65"), ("mape", "    40.3")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    0.94"), ("mape", "    22.6")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    0.98"), ("mape", "   126.7")]),
                "HSG": OrderedDict([("mae", "    0.18"), ("mape", "    15.1")]),
                "DB4": OrderedDict([("mae", "    0.70"), ("mape", "    54.8")]),
            },
        },
        "CCSD-CP-atz": {
            "hb": {
                "S22": OrderedDict([("mae", "    1.41"), ("mape", "    10.5")]),
                "HBC1": OrderedDict([("mae", "    1.11"), ("mape", "     8.4")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    1.39"), ("mape", "     9.0")]),
                "DB4": OrderedDict([("mae", "    1.30"), ("mape", "     9.3")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    1.27"), ("mape", "    23.5")]),
                "HBC1": OrderedDict([("mae", "    1.11"), ("mape", "     8.4")]),
                "NBC1": OrderedDict([("mae", "    0.97"), ("mape", "   123.2")]),
                "HSG": OrderedDict([("mae", "    0.77"), ("mape", "    57.9")]),
                "DB4": OrderedDict([("mae", "    1.03"), ("mape", "    53.3")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    1.20"), ("mape", "    29.6")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    0.97"), ("mape", "   123.2")]),
                "HSG": OrderedDict([("mae", "    0.67"), ("mape", "    66.1")]),
                "DB4": OrderedDict([("mae", "    0.95"), ("mape", "    73.0")]),
            },
        },
        "CCSD-CP-adtz": {
            "hb": {
                "S22": OrderedDict([("mae", "    1.03"), ("mape", "     7.6")]),
                "HBC1": OrderedDict([("mae", "    0.76"), ("mape", "     5.9")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    1.02"), ("mape", "     6.6")]),
                "DB4": OrderedDict([("mae", "    0.94"), ("mape", "     6.7")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    1.05"), ("mape", "    20.0")]),
                "HBC1": OrderedDict([("mae", "    0.76"), ("mape", "     5.9")]),
                "NBC1": OrderedDict([("mae", "    0.90"), ("mape", "   113.9")]),
                "HSG": OrderedDict([("mae", "    0.62"), ("mape", "    46.7")]),
                "DB4": OrderedDict([("mae", "    0.83"), ("mape", "    46.6")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    1.06"), ("mape", "    25.7")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    0.90"), ("mape", "   113.9")]),
                "HSG": OrderedDict([("mae", "    0.55"), ("mape", "    53.3")]),
                "DB4": OrderedDict([("mae", "    0.84"), ("mape", "    64.3")]),
            },
        },
        "MP2-CP-adz": {
            "hb": {
                "S22": OrderedDict([("mae", "    1.64"), ("mape", "    12.2")]),
                "HBC1": OrderedDict([("mae", "    1.41"), ("mape", "    10.4")]),
                "NBC1": None,
                "HSG": OrderedDict([("mae", "    1.42"), ("mape", "     9.1")]),
                "DB4": OrderedDict([("mae", "    1.49"), ("mape", "    10.6")]),
            },
            "default": {
                "S22": OrderedDict([("mae", "    0.97"), ("mape", "    16.2")]),
                "HBC1": OrderedDict([("mae", "    1.41"), ("mape", "    10.4")]),
                "NBC1": OrderedDict([("mae", "    0.70"), ("mape", "    95.9")]),
                "HSG": OrderedDict([("mae", "    0.46"), ("mape", "    30.5")]),
                "DB4": OrderedDict([("mae", "    0.88"), ("mape", "    38.3")]),
            },
            "mxdd": {
                "S22": OrderedDict([("mae", "    0.65"), ("mape", "    18.0")]),
                "HBC1": None,
                "NBC1": OrderedDict([("mae", "    0.70"), ("mape", "    95.9")]),
                "HSG": OrderedDict([("mae", "    0.30"), ("mape", "    34.1")]),
                "DB4": OrderedDict([("mae", "    0.55"), ("mape", "    49.3")]),
            },
        },
    }
    # print serrors['CCSD-CP-adtz']['hb']['HSG']['mape']

    columnplan = [
        ["l", r"""Method \& Basis Set""", "", label, {}],
        ["d", r"S22", "HB", val, {"sset": "hb", "dbse": "S22"}],
        ["d", r"S22", "MX/DD", val, {"sset": "mxdd", "dbse": "S22"}],
        ["d", r"S22", "TT", val, {"sset": "default", "dbse": "S22"}],
        ["d", r"Overall", "HB", val, {"sset": "hb", "dbse": "DB4"}],
        ["d", r"Overall", "MX/DD", val, {"sset": "mxdd", "dbse": "DB4", "err": "mape"}],
        ["d", r"Overall", "TT", val, {"sset": "default", "dbse": "DB4"}],
    ]

    tbl, ind = table_generic(
        columnplan=columnplan,
        dbse=["DB4"],
        serrors=serrors,
        mtd=["MP2", "CCSD"],
        bas=["adz", "atz"],
        opt=["CP"],
        err=["mae", "mape"],
        theme="test",
        subjoin=False,
    )
    print("\n".join(tbl))



from cdsg_plot.qcdb_plot import threads as mpl_threads


def plotly_threads(data, labels, color=None, title='', xlimit=4.0, xlimitleft=None, xticks=None,
    mae=None, mape=None,
    mousetext=None, mouselink=None, mouseimag=None, mousetitle=None, mousediv=None,
    labeled=True, view=True,
    saveas=None, relpath=False, graphicsformat=['pdf']):
    """Generates a tiered slat diagram between model chemistries with
    errors (or simply values) in list *data*, which is supplied as part of the
    dictionary for each participating reaction, along with *dbse* and *rxn* keys
    in argument *data*. The plot is labeled with *title* and each tier with
    an element of *labels* and plotted at *xlimit* from the zero-line. If
    *color* is None, slats are black, if 'sapt', colors are taken from *color*
    key in *data* [0, 1]. Summary statistics *mae* are plotted on the
    overbound side and relative statistics *mape* on the underbound side.
    HTML code for mouseover if mousetext or mouselink or mouseimag specified
    based on recipe of Andrew Dalke from
    http://www.dalkescientific.com/writings/diary/archive/2005/04/24/interactive_html.html
    """
    import hashlib

    # initialize tiers/wefts
    Nweft = len(labels)
    lenS = 0.2
    gapT = 0.04
    positions = range(-1, -1 * Nweft - 1, -1)
    posnS = []
    for weft in range(Nweft):
        posnS.extend([positions[weft] + lenS, positions[weft] - lenS, None])
    posnT = []
    for weft in range(Nweft - 1):
        posnT.extend([positions[weft] - lenS - gapT, positions[weft + 1] + lenS + gapT, None])
    posnM = []
    if xlimitleft is None:
        xlimitleft = -1 * xlimit
    xrange = xlimit - xlimitleft
    if xticks is None:
        xticks = [xlimitleft + 0.25 * xrange, xlimitleft + 0.375 * xrange,
                  xlimitleft + 0.5 * xrange, xlimitleft + 0.625 * xrange, xlimitleft + 0.75 * xrange]

    # initialize plot
    import plotly.graph_objects as go
    fig = go.Figure()

    fig.update_layout(
        autosize=False,
        width=72 * 11,
        height=72 * Nweft * 0.8,
        margin=dict(b=36, l=7, r=7, t=34, pad=0),
        showlegend=False,
        xaxis=dict(range=[xlimitleft, xlimit], tickvals=xticks, zeroline=True, zerolinewidth=3),
        yaxis=dict(range=[-1 * Nweft - 1, 0], showticklabels=False),
    )

    # label plot and tiers
    annot = []
    annot.append(go.layout.Annotation(
        x=xlimitleft + 0.05 * xrange,
        y=-0.25,
        align='left',
        #xanchor='left',
        text=title,
        showarrow=False,
        font=dict(size=12),
    ))
    for weft in labels:
        annot.append(go.layout.Annotation(
            x=xlimitleft + 0.05 * xrange,
            y=-(1.0 + labels.index(weft)),
            xref="x",
            yref="y",
            text=weft,
            align='left',
            showarrow=False,
            font=dict(size=18),
        ))
    fig.update_layout(annotations=annot)

#     if labeled:
#         ax.text(-0.9 * xlimit, -0.25, title,
#             verticalalignment='bottom', horizontalalignment='left',
#             family='Times New Roman', weight='bold', fontsize=12)
#         for weft in labels:
#             ax.text(-0.9 * xlimit, -(1.2 + labels.index(weft)), weft,
#                 verticalalignment='bottom', horizontalalignment='left',
#                 family='Times New Roman', weight='bold', fontsize=18)

    # plot reaction errors and threads
    for rxn in data:

        # preparation
        xvals = rxn['data']
        clr = rxn['color'] if 'color' in rxn else 'green' 
        slat = []
        for weft in range(Nweft):
            slat.extend([xvals[weft], xvals[weft], None])
        thread = []
        for weft in range(Nweft - 1):
            thread.extend([xvals[weft], xvals[weft + 1], None])

        # plotting
        fig.add_trace(go.Scatter(x=slat,
                                 y=posnS,
                                 mode='lines',
                                 name=rxn['sys'],
                                 line=dict(
                                     color=clr,
                                     dash='solid',
                                     width=1.0,
                                 ),
                                ))

        fig.add_trace(go.Scatter(x=thread, y=posnT,
                                 mode='lines',
                                 name=rxn['sys'],
                                 opacity=0.6, #0.3,
                                 showlegend=False,
                                 line=dict(
                                     color=clr,
                                     dash='solid',
                                     width=0.5,
                                 ),                        
                                ))

    # plot trimmings
    if mae is not None:
        fig.add_trace(go.Scatter(x=[-x for x in mae], y=list(positions),
                                 mode='markers',
                                 name='MAE',
                                 marker=dict(
                                     color='black',
                                     symbol='square'),
                                ))
    if mape is not None:  # equivalent to MAE for a 10 kcal/mol IE
        fig.add_trace(go.Scatter(x=[0.025 * x for x in mape], y=list(positions),
                                 mode='markers',
                                 name='MA%E',
                                 marker=dict(
                                     color='black',
                                     symbol='circle'),
                                ))

    # save and show
    pltuid = title + '_' + ('lbld' if labeled else 'bare') + '_' + hashlib.sha1((title + repr(labels) + repr(xlimit)).encode()).hexdigest()

    if view:
        fig.show()
    return fig


if __name__ == "__main__":

    merge_dats = [
        {
            "show": "a",
            "db": "HSG",
            "sys": "1",
            "data": [0.3508, 0.1234, 0.0364, 0.0731, 0.0388],
        },
        {
            "show": "b",
            "db": "HSG",
            "sys": "3",
            "data": [0.2036, -0.0736, -0.1650, -0.1380, -0.1806],
        },
        # {'show':'', 'db':'S22', 'sys':'14', 'data':[np.nan, -3.2144, np.nan, np.nan, np.nan]},
        {
            "show": "c",
            "db": "S22",
            "sys": "14",
            "data": [None, -3.2144, None, None, None],
        },
        {
            "show": "d",
            "db": "S22",
            "sys": "15",
            "data": [-1.5090, -2.5263, -2.9452, -2.8633, -3.1059],
        },
        {
            "show": "e",
            "db": "S22",
            "sys": "22",
            "data": [0.3046, -0.2632, -0.5070, -0.4925, -0.6359],
        },
    ]

    mpl_threads_plt = mpl_threads(
        merge_dats,
        labels=["d", "t", "dt", "q", "tq"],
        color="sapt",
        title="MP2-CPa[]z",
        mae=[0.25, 0.5, 0.5, 0.3, 1.0],
        mape=[20.1, 25, 15, 5.5, 3.6],
        graphicsformat=["png"],
        view=False,
    )

    ply_threads_plt = plotly_threads(
        merge_dats,
        labels=["d", "t", "dt", "q", "tq"],
        color="sapt",
        title="MP2-CPa[]z",
        mae=[0.25, 0.5, 0.5, 0.3, 1.0],
        mape=[20.1, 25, 15, 5.5, 3.6],
    )

